/*
 * Copyright (c) 2011, Stefan Götz <stefan.goetz@web.de>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

/**
 * @internal
 * @file
 * RList implementation of doubly-linked lists.
 */

#include <stdlib.h>
#include <assert.h>

#include "rlist/double.h"

bool drl_check_list_(const struct drl_list *const list)
{
    /* is the list a valid object? */
    assert(list);
    /* if either head or tail is NULL, the other one has to be NULL, too */
    assert((list->head && list->tail) || (!list->head && !list->tail));

    /* if node_count is 0, then the head and tail pointers must be NULL */
    assert(list->node_count != 0 || (!list->head && !list->tail));
    /* if head and tail are NULL, node_count must be 0 */
    assert((list->head && list->tail) || list->node_count == 0);

    /* if node_count is > 0, then the head and tail pointers must not be NULL */
    assert(list->node_count == 0 || (list->head && list->tail));
    /* if head and tail are not NULL, then node_count must be > 0 */
    assert((!list->head && !list->tail) || list->node_count > 0);

    /* if there is a head element, its prev pointer must be NULL */
    assert(!list->head || !list->head->prev);
    /* if there is a tail element, its next pointer must be NULL */
    assert(!list->tail || !list->tail->next);

    /* if node_count is 1, then head and tail must point to the same element and their next and prev pointers must match up */
    assert(list->node_count != 1 || (list->head == list->tail && !list->head->next && !list->tail->prev));
    /* only if there is exactly one node in the list, node_count must be 1 */
    assert(!list->head || !list->tail || list->head != list->tail || list->head->next || list->tail->prev || list->node_count == 1);

    /* if node_count is > 1, then head->next and tail->prev must not be NULL */
    assert(list->node_count <= 1 || (list->head->next && list->tail->prev));
    /* if head->next and tail->prev are not NULL, then node_count must be > 1 */
    assert(!list->head || !list->tail || (!list->head->next && !list->tail->prev) || list->node_count > 1);

    /* if node_count is 2, then head->next must be tail and tail->prev must be head */
    assert(list->node_count != 2 || (list->head->next == list->tail && list->tail->prev == list->head));
    /* if head->next is tail and tail->prev is head, then node_count must be 2 */
    assert(!list->head || !list->tail || !list->head->next || !list->tail->prev || list->head->next != list->tail || list->tail->prev != list->head || list->node_count == 2);

    /* if node_count is 3, then head->next must be equal to tail->prev */
    assert(list->node_count != 3 || (list->head->next == list->tail->prev));
    /* if head->next is equal to tail->prev, then node_count must be 3 */
    assert(!list->head || !list->tail || !list->head->next || !list->tail->prev || list->head->next != list->tail->prev || list->node_count == 3);

    /* if node_count is > 3, then head->next must not be equal to tail->prev */
    assert(list->node_count <= 3 || (list->head->next != list->tail->prev));
    /* if head->next is not equal to tail->prev, then node_count must be > 3 */
    assert(!list->head || !list->tail || !list->head->next || !list->tail->prev || list->head->next == list->tail || list->tail->prev == list->head || list->head->next == list->tail->prev || list->node_count > 3);

    return list != NULL;
}

bool drl_check_iter_in_list_(const struct drl_iterator *const iter)
{
    /* is the iterator a valid object? */
    assert(iter);
    /* is the iterator's list consistent? */
    drl_check_list_(iter->list);
    /* is the iterator in the correct state? */
    assert(iter->state == RL_ISTATE_IN_LIST_);
    /* is the current node a valid object? */
    assert(iter->node);
    /* is the list non-empty so that this iterator can, in fact, be 'in the list'? */
    assert(iter->list->node_count > 0);

    return iter != NULL;
}

bool drl_check_iter_post_delete_(const struct drl_iterator *const iter)
{
    /* is the iterator a valid object? */
    assert(iter != NULL);
    /* is the iterator's list consistent? */
    drl_check_list_(iter->list);
    /* is the iterator in the correct state? */
    assert(iter->state == RL_ISTATE_POST_DELETE_);

    return iter != NULL;
}

bool drl_check_iter_end_of_list_(const struct drl_iterator *const iter)
{
    /* is the iterator a valid object? */
    assert(iter != NULL);
    /* is the iterator's list consistent? */
    drl_check_list_(iter->list);
    /* is the iterator in the correct state? */
    assert(iter->state == RL_ISTATE_END_OF_LIST_);
    /* is the current node NULL to indicate EOL? */
    assert(iter->node == NULL);

    return iter != NULL;
}

bool drl_check_iter_in_list_or_end_of_list_(const struct drl_iterator *const iter)
{
    assert(iter);
    assert((iter->state == RL_ISTATE_IN_LIST_  && drl_check_iter_in_list_(iter)) ||
           (iter->state == RL_ISTATE_END_OF_LIST_ && drl_check_iter_end_of_list_(iter)));

    return iter != NULL;
}

bool drl_insert_before(struct drl_iterator *const iter,
                       void *const payload)
{
    assert(drl_check_iter_in_list_or_end_of_list_(iter));

    struct drl_node_ *const new_node = malloc(sizeof(*new_node));
    if (new_node) {
        struct drl_node_ *const prev = iter->node ? iter->node->prev : iter->list->tail;
        struct drl_node_ *const next = iter->node;

        *new_node = (struct drl_node_) { payload, next, prev };

        if (prev) {
            prev->next = new_node;
        }
        if (next) {
            next->prev = new_node;
        }

        if (iter->list->head == iter->node) {
            iter->list->head = new_node;
        }
        if (iter->node == NULL) {
            iter->list->tail = new_node;
        }
        iter->list->node_count += 1;

        assert(drl_check_iter_in_list_or_end_of_list_(iter));

        return true;
    }

    return false;
}

void *drl_del(struct drl_iterator *const iter)
{
    void *ret = NULL;
    assert(drl_check_iter_in_list_(iter));

    struct drl_node_ *const victim = iter->node;
    assert(victim);

    /* the following pointers may refer to the victim node:
     * 1) list->head
     * 2) list->tail
     * 3) victim->prev->next
     * 4) victim->next->prev
     * 5) iter->node */

    /* 1 of 5 */
    if (iter->list->head == victim) {
        iter->list->head = victim->next;
    }
    /* 2 of 5 */
    if (iter->list->tail == victim) {
        iter->list->tail = victim->prev;
    }
    /* 3 of 5 */
    if (victim->prev) {
		victim->prev->next = victim->next;
	}
	/* 4 of 5 */
	if (victim->next) {
		victim->next->prev = victim->prev;
	}
	/* 5 of 5 */
    /* Update the iterator's node pointer and set its state so that drl_next() or drl_prev() fixes it up */
    iter->node = victim->next;
    iter->state = RL_ISTATE_POST_DELETE_;

    iter->list->node_count -= 1;

    /* free payload and node memory */
    if (iter->list->free_fn) {
        iter->list->free_fn(victim->payload);
    } else {
        ret = victim->payload;
    }
    free(victim);

    assert(drl_check_iter_post_delete_(iter));

    return ret;
}

void drl_del_all(struct drl_list *const list)
{
    struct drl_iterator i;

    assert(drl_check_list_(list));

    while (!drl_is_eol(drl_first(list, &i))) {
        drl_del(&i);
    }

    assert(drl_check_list_(list));
}

/**
 * @internal
 * A rl_map_fn() call-back function which copies the iterator of the first list node it encounters to @a context and terminates the drl_map() iteration.
 *
 * @param iter points to an iterator that refers to a list node.
 * @param context assumed to point to an drl_iterator object.
 *  After calling this function, the data in @a iter is copied to @a context.
 * @return Returns false.
 */
static int rl_map_return_first_match_(__attribute__ ((unused)) struct drl_iterator *const iter,
                                      __attribute__ ((unused)) void *const context)
{
    return 1;
}

struct drl_iterator *drl_find(struct drl_iterator *const iter,
                              rl_match_fn match_fn,
                              void *const match_context)
{
    assert(drl_check_iter_in_list_or_end_of_list_(iter));
    assert(match_fn);

    drl_map(iter, rl_map_return_first_match_, NULL, match_fn, match_context);

    assert(drl_check_iter_in_list_or_end_of_list_(iter));

    return iter;
}

/**
 * @internal
 * Match a node by its payload.
 *
 * @param payload points to the payload of the node to check.
 * @param context points to the payload to be found in a node.
 * @return true when @a payload and @a context are equal.
 */
static bool rl_match_payload_(void *const payload,
                              void *const context)
{
    return payload == context;
}

struct drl_iterator *drl_find_payload(struct drl_iterator *const iter,
                                      void *const payload)
{
    return drl_find(iter, rl_match_payload_, payload);
}

int drl_map(struct drl_iterator *const iter,
            drl_map_fn map_fn,
            void *const map_context,
            rl_match_fn match_fn,
            void *const match_context)
{
    int ret = 0;

    assert(drl_check_iter_in_list_or_end_of_list_(iter));
    assert(map_fn);

    for (; !drl_is_eol(iter); drl_next(iter)) {
        if (!match_fn || match_fn(drl_payload(iter), match_context)) {
            ret = map_fn(iter, map_context);
            if (ret != 0) {
                break;
            }
        }
    }

    assert(drl_check_iter_in_list_or_end_of_list_(iter));

    return ret;
}
