/*
 * Copyright (c) 2011, Stefan Götz <stefan.goetz@web.de>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

/**
 * @file
 * Public declarations common to multiple RList data structures.
 */

#ifndef RLIST_COMMON_H
#define RLIST_COMMON_H

#include <stdbool.h>

/**
 * A function type for deleting the payload of a list node.
 * A function of this type can be associated with a list to automatically delete the payload of a node whenever a node is deleted.
 *
 * This helps in avoiding memory leaks:
 * When deleting a list node, it is easy to forget to de-allocate its payload.
 * When a list is associated with a rl_free_fn() call-back function, the call-back is invoked whenever a node is deleted from the list.
 * This is primarily useful when the list payload is allocated dynamically and the list serves as the primary container for the payload (i.e., there is no other data structure that points to the same payload data).
 * @sa srl_init(), drl_init()
 */
typedef void (*rl_free_fn)(void *payload);

/**
 * A type for call-back functions used to match list nodes against custom criteria, e.g., for finding list nodes with specific payload.
 *
 * Typically, call-backs of this type are used to linearly iterate over a list and either stop at the first @e match or to apply some function to all @e matching nodes.
 * It it is recommended to make such call-back functions efficient because they are applied to potentially large numbers of list nodes.
 *
 * Using call-back functions for these operations gives the user of drl_list full customizability on the matching criteria while the generic iteration code is provided as an RList function and does not need to be re-implemented and customized over and over.
 *
 * @param payload points to the payload of a list node that is to be classified as a @e match or not.
 * @param context points to a custom memory location that the call-back function requires as input or for storing data across calls.
 *  @a context may be used, e.g., for storing known payload data at the beginning of a search so the call-back only needs to compare @a context and @a payload to determine a @e match.
 *  Another use is to read and update changing information with every call, such as a position counter.
 * @return The call-back function is expected to return true if the list node with @a payload is to be considered a @e match.
 *  It is expected to return false if the list node is not to be considered a @e match.
 */
typedef bool (*rl_match_fn)(void *const payload,
                            void *const context);

/**
 * @internal
 * @sa srl_iterator, drl_iterator
 */
enum rl_iterator_state_ {
    /* in-list, iterator can be used with any function */
    RL_ISTATE_IN_LIST_,
    /* post-delete, do not use iterator with any public function except
     * drl_next() or drl_prev() */
    RL_ISTATE_POST_DELETE_,
    /* end-of-list, do not use iterator with any public function except
     * drl_insert_before() */
    RL_ISTATE_END_OF_LIST_,
};

#endif /* RLIST_COMMON_H */
