/*
 * Copyright (c) 2011, Stefan Götz <stefan.goetz@web.de>
 *
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

/**
 * @file
 * Public RList API for doubly linked lists.
 *
 * @internal
 * The internal data structures of a doubly-linked list:
 * <pre>
 * +------------+
 * |drl_iterator|------------------------------------+
 * +------------+                                    |
 *       |                                           |
 *       | list                                      | node
 *       v          head                 next        v
 * +------------+ -------> +---------+ -------> +---------+ -- ... --> NULL
 * | drl_list   |          |drl_node_|   prev   |drl_node_|
 * +------------+       +- +---------+ <------- +---------+ <- ...
 *       |         prev |       |                    |          *
 *       |      NULL <--+       | payload            | payload  |
 *       |                      v                    v          |
 *       |                 +---------+          +---------+     |
 *       |                 | payload |          | payload |     |
 *       |                 +---------+          +---------+     |
 *       |          tail                                        |
 *       +------------------------------------------------------+
 * </pre>
 */

#ifndef RLIST_DOUBLE_H
#define RLIST_DOUBLE_H

#include <stddef.h>
#include <stdbool.h>
#include <assert.h>

#include "rlist/platform.h"
#include "rlist/common.h"

C_DECLARATIONS_START

/**
 * An initializer constant for drl_list objects.
 *
 * After assigning this value to a list, it is empty and not associated with an rl_free_fn() function.
 *
 * Static or local declarations of drl_list objects should be initialized in the declaration with this value (e.g., <tt>static struct drl_list list = DRL_INIT;</tt>).
 * The drl_init() function should be used where initialization at compile time is not possible so it has to occur at runtime, e.g., for dynamically allocated drl_list objects.
 */
#define DRL_INIT { 0, NULL, NULL, NULL }

/**
 * An initializer constant for drl_list objects with a custom call-back function for de-allocating payload data.
 *
 * After assigning this value to a list, it is empty and has the specified rl_free_fn() function associated.
 *
 * Static or local declarations of drl_list objects should be initialized in the declaration with this value (e.g., <tt>static struct drl_list list = DRL_INIT;</tt>).
 * The drl_init() function should be used where initialization at compile time is not possible so it has to occur at runtime, e.g., for dynamically allocated drl_list objects.
 *
 * @param free_fn points to a function with the type rl_free_fn() for de-allocating payload data.
 */
#define DRL_INIT_WITH(free_fn) { 0, free_fn, NULL, NULL }

/**
 * @internal
 * A single list node that refers to other nodes and its payload.
 *
 * This type is internal to the list implementation.
 * Users of drl_list should never use this type or one of its instances.
 */
struct drl_node_ {
    /* The payload data provided by the user */
    void *payload;
    /* The next node in the list.
     * If next is NULL, this node is the last node in the list. */
    struct drl_node_ *next;
    /* The previous node in the list.
     * If prev is NULL, this node is the first node in the list. */
    struct drl_node_ *prev;
};

/**
 * A doubly linked list object.
 *
 * drl_list objects represent a complete list, through which the list nodes can be accessed.
 * drl_list objects are typically declared statically or as function variables whereas it is less common to allocated them dynamically.
 * drl_list objects @b must be initialized (either via DRL_INIT or via drl_init()) before being used or passed to an RList function.
 * This type is intended to be completely opaque to users of RList.
 * Do not access any of its fields directly, as they might change any time.
 * Use the RList accessor functions instead.
 */
struct drl_list {
    /**
     * @internal
     * Stores the number of nodes currently in the list.
     * It is updated by drl_insert() and drl_del().
     */
    unsigned int node_count;
    /**
     * @internal
     * Points to a function for freeing node payload data.
     * If it is not NULL, it is invoked as a call-back function by
     * drl_del().
     */
    rl_free_fn free_fn;
    /**
     * @internal
     * Points to the first list node.
     * If head is NULL, the list is empty and tail should be NULL and node_count should be 0.
     */
    struct drl_node_ *head;
    /**
     * @internal
     * Points to the last list node.
     * If tail is NULL, the list is empty and head should be NULL and node_count should be 0.
     */
    struct drl_node_ *tail;
};

/**
 * A type for representing a node or a position in a doubly linked list.
 *
 * This is the main type which users of RList interact with.
 * It is the basis for most list operations, such as iteration, deletion, or searching.
 *
 * This type is intended to be completely opaque to users of RList.
 * Do not access any of its fields directly, as they might change any time.
 * Use the RList accessor functions instead.
 *
 * @internal
 * An iterator can have three states:
 * in-list (IL), end-of-list (EOL), post-delete (PD).
 *
 * By default, iterators are in-list and node is not NULL.
 *
 * An EOL iterator indicates that the beginning or end of the list has been reached.
 * Having a valid iterator object that can signify this state is more convenient than using, e.g., NULL pointers for EOL iterators.
 * The EOL state implies that node == NULL.
 *
 * Calling drl_del() on an iterator puts it into the post-delete state.
 * This allows to mix deletion and iteration.
 * In a straightforward implementation, deleting a node via an iterator would implicitly advance the iterator to the next node.
 * Thus, code that mixes deletion and iteration would need to take into account that both drl_del() and drl_next()/drl_prev() modify an iterator's position, which is cumbersome.
 * We thus move this case distinction into the RList code so drl_next() and drl_prev() can (or rather must) be called after drl_del() to advance the iterator and put it into the IL or EOL state.
 */
struct drl_iterator {
    /**
     * @internal
     * The complete list this iterator operates on.
     * It is required to determine the list head and the associated
     * rl_free_fn call-back function.
     */
    struct drl_list *list;
    /**
     * @internal
     * The current node this iterator refers to.
     */
    struct drl_node_ *node;
    /**
     * @internal
     * The current state of this iterator.
     * This state depends on where the iterator points to in the list and whether the current list item was deleted or not.
     * The state is updated by the iteration functions drl_next() and drl_prev().
     */
    enum rl_iterator_state_ state;
};

/**
 * A type for call-back functions that are applied to multiple list nodes.
 *
 * Typically, call-backs of this type are used to apply a certain operation (such as modifying the payload or deleting nodes) to some or all nodes in a list (@sa drl_map()).
 * It is recommended to make such call-back functions efficient because they are applied to potentially large numbers of list nodes.
 *
 * Using call-back functions for these operations gives the user of drl_list full customizability on the action performed on the list nodes while the generic iteration code is provided as an RList function and does not need to be re-implemented and customized over and over.
 *
 * @param iter points to an iterator that refers to a list node.
 *  The call-back function may freely manipulate the node's payload.
 *  It may also freely manipulate the iterator via any of the RList functions, including deleting the current node.
 * @param context points to a custom memory location that the call-back function requires as input or for storing data across calls.
 * @return If the call-back function returns 0, the iteration over the list continues and the call-back function is invoked for the remaining list nodes.
 *  If a non-zero value is returned, the iteration stops and the call-back is not invoked for any further list nodes.
 */
typedef int (*drl_map_fn)(struct drl_iterator *const iter,
                          void *const context);

/* The following functions are required by some inline functions here in the
 * header file.
 * Since they are only called from assert(), i.e. not present in production code, the performance impact of not inlining them, too, can be ignored. */
/**
 * @internal
 * Do a sanity check to determine whether a given list object contains consistent data.
 *
 * @param list Points to the list object to check.
 *  Calling this function with an inconsistent list triggers an assertion.
 * @return true
 */
bool drl_check_list_ ARGS((const struct drl_list *const list));

/**
 * @internal
 * Check whether an iterator in the IL state is internally consistent.
 *
 * @sa drl_iterator, drl_check_iter_end_of_list_(), drl_check_iter_post_delete_(), drl_check_iter_in_list_()
 *
 * @param iter Points to the iterator object to check.
 *  Failing the sanity checks triggers an assertion.
 * @return true
 */
bool drl_check_iter_in_list_ ARGS((const struct drl_iterator *const iter));

/**
 * @internal
 * Check whether an iterator in the PD state is internally consistent.
 *
 * @sa drl_iterator, drl_check_iter_in_list_(), drl_check_iter_end_of_list_(), drl_check_iter_in_list_or_end_of_list_()
 *
 * @param iter Points to the iterator object to check.
 *  Failing the sanity checks triggers an assertion.
 * @return true
 */
bool drl_check_iter_post_delete_ ARGS((const struct drl_iterator *const iter));

/**
 * @internal
 * Check whether an iterator in the EOL state is internally consistent.
 *
 * @sa drl_iterator, drl_check_iter_in_list_(), drl_check_iter_post_delete_(), drl_check_iter_in_list_or_end_of_list_()
 *
 * @param iter Points to the iterator object to check.
 *  Failing the sanity checks triggers an assertion.
 * @return true
 */
bool drl_check_iter_end_of_list_ ARGS((const struct drl_iterator *const iter));

/**
 * @internal
 * Check whether an iterator is internally consistent and either in the IL or the EOL state.
 *
 * @sa drl_iterator, enum drl_iterator_state_, drl_check_iter_in_list_(), drl_check_iter_post_delete_(), drl_check_iter_in_list_or_end_of_list_()
 *
 * @param iter Points to the iterator object to check.
 *  Failing the sanity checks triggers an assertion.
 * @return true
 */
bool drl_check_iter_in_list_or_end_of_list_ ARGS((const struct drl_iterator *const iter));

/**
 * Retrieve the payload an iterator object is associated with.
 *
 * The runtime complexity of this function is O(1).
 *
 * @param iter points to the iterator for which to retrieve the payload pointer.
 *  The result of calling this function with <tt>iter == NULL</tt> is undefined.
 *  The result of calling this function with an inconsistent iterator (@sa drl_del()) is undefined.
 * @return Returns the payload of the list node that @a iter refers to.
 */
static inline void *drl_payload(const struct drl_iterator *const iter)
{
    assert(iter);
    return iter->node->payload;
}

/**
 * Initialize a drl_list object at runtime.
 *
 * This function should be used where initialization at compile time is not possible so it has to occur at runtime, e.g., for dynamically allocated drl_list objects.
 * Static or local declarations of drl_list objects should be initialized in their declaration via DRL_INIT.
 *
 * The runtime complexity of this function is O(1).
 *
 * @param list points to the list object to initialize.
 *  The result of calling this function with <tt>list == NULL</tt> is undefined.
 * @param free_fn points to a rl_free_fn() call-back function for de-allocating payload data.
 *  If @a free_fn is @c NULL, the list does not automatically de-allocate payload data.
 *  Proper memory management of the payload data is then the responsibility of the user of @a list.
 */
static inline void drl_init(struct drl_list *const list,
                            rl_free_fn free_fn)
{
    assert(list);
    *list = (struct drl_list) DRL_INIT_WITH(free_fn);
    assert(drl_check_list_(list));
}

/**
 * Retrieve the number of nodes currently held by a list.
 *
 * The runtime complexity of this function is O(1).
 *
 * @param list points to the list object of which to retrieve the node count.
 *  The result of calling this function with <tt>list == NULL</tt> is undefined.
 * @return The number of nodes currently held by @a list.
 */
static inline unsigned int drl_size(const struct drl_list *const list)
{
    assert(drl_check_list_(list));
    return list->node_count;
}

/**
 * Initialize an iterator object so it refers to the first node in a list.
 *
 * If the list is empty, the iterator is updated so that drl_is_eol() returns true.
 *
 * The runtime complexity of this function is O(1).
 *
 * @param list Points to the list from which to retrieve the first node.
 *  The result of calling this function with <tt>list == NULL</tt> is undefined.
 * @param iter Points to the iterator object to initialize.
 *  The result of calling this function with <tt>iter == NULL</tt> is undefined.
 * @return This function returns @a iter.
 */
static inline struct drl_iterator *drl_first(struct drl_list *const list,
                                             struct drl_iterator *const iter)
{
    assert(drl_check_list_(list));
    assert(iter);

    *iter = (struct drl_iterator) { list, list->head, list->head ? RL_ISTATE_IN_LIST_ : RL_ISTATE_END_OF_LIST_ };

    return iter;
}

/**
 * Initialize an iterator object so it refers to the last node in a list.
 *
 * If the list is empty, the iterator is updated so that drl_is_eol() returns true.
 *
 * The runtime complexity of this function is O(1).
 *
 * @param list Points to the list from which to retrieve the last node.
 *  The result of calling this function with <tt>list == NULL</tt> is undefined.
 * @param iter Points to the iterator object to initialize.
 *  The result of calling this function with <tt>iter == NULL</tt> is undefined.
 * @return This function returns @a iter.
 */
static inline struct drl_iterator *drl_last(struct drl_list *const list,
                                            struct drl_iterator *const iter)
{
    assert(drl_check_list_(list));
    assert(iter);

    *iter = (struct drl_iterator) { list, list->tail, list->tail ? RL_ISTATE_IN_LIST_ : RL_ISTATE_END_OF_LIST_ };

    return iter;
}

/**
 * Update an iterator object so it refers to the 'next' list node, i.e., the node following the one the iterator refers to before calling this function.
 *
 * After passing an iterator to drl_del(), you must call drl_next() or drl_prev() on the same iterator before passing it to any other RList function.
 * This advances the iterator to the node after or before the deleted node.
 *
 * This function is primarily intended for iterating through a list.
 *
 * The runtime complexity of this function is O(1).
 *
 * @param iter Points to an iterator object.
 *  The iterator may not be uninitialized and must have previously been returned by one of the RList functions.
 *  If @a iter refers to either end of the list, i.e., if <tt>drl_is_eol(iter)</tt> returns true, this function does not modify @a iter.
 *  The result of calling this function with <tt>iter == NULL</tt> is undefined.
 * @return This function returns @a iter.
 */
static inline struct drl_iterator *drl_next ARGS((struct drl_iterator *const iter))
{
    assert(iter);

    switch (iter->state) {
    case RL_ISTATE_IN_LIST_:
        assert(drl_check_iter_in_list_(iter));
        iter->node = iter->node->next;
        break;
    case RL_ISTATE_POST_DELETE_:
        assert(drl_check_iter_post_delete_(iter));
        break;
    case RL_ISTATE_END_OF_LIST_:
        assert(drl_check_iter_end_of_list_(iter));
        break;
    }
    
    if (iter->node) {
        iter->state = RL_ISTATE_IN_LIST_;
    } else {
        iter->state = RL_ISTATE_END_OF_LIST_;
    }

    assert(drl_check_iter_in_list_or_end_of_list_(iter));

    return iter;
}

/**
 * Update an iterator object so it refers to the 'previous' list node, i.e., the node preceeding the one the iterator refers to before calling this function.
 *
 * After passing an iterator to drl_del(), you must call drl_next() or drl_prev() on the same iterator before passing it to any other RList function.
 * This advances the iterator to the node after or before the deleted node.
 *
 * This function is primarily intended for iterating through a list.
 *
 * The runtime complexity of this function is O(1).
 *
 * @param iter Points to an iterator object.
 *  The iterator may not be uninitialized and must have previously been returned by one of the RList functions.
 *  If @a iter refers to either end of the list, i.e., if <tt>drl_is_eol(iter)</tt> returns true, this function does not modify @a iter.
 *  The result of calling this function with <tt>iter == NULL</tt> is undefined.
 * @return This function returns @a iter.
 */
static inline struct drl_iterator *drl_prev ARGS((struct drl_iterator *const iter))
{
    assert(iter);

    switch (iter->state) {
    case RL_ISTATE_IN_LIST_:
        assert(drl_check_iter_in_list_(iter));
        iter->node = iter->node->prev;
        break;
    case RL_ISTATE_POST_DELETE_:
        assert(drl_check_iter_post_delete_(iter));
        break;
    case RL_ISTATE_END_OF_LIST_:
        assert(drl_check_iter_end_of_list_(iter));
        break;
    }
    
    if (iter->node) {
        iter->state = RL_ISTATE_IN_LIST_;
    } else {
        iter->state = RL_ISTATE_END_OF_LIST_;
    }

    assert(drl_check_iter_in_list_or_end_of_list_(iter));

    return iter;
}

/**
 * Determine whether an iterator refers to either end of a list, i.e., whether it iterated beyond the last list node or before the first list node.
 *
 * This is primarily useful to determine when an iteration over a list has ended.
 * In this state, the iterator does not refer to a valid list node and should not be passed to any RList functions except drl_is_eol().
 *
 * The runtime complexity of this function is O(1).
 *
 * @param iter Points to the iterator object to check.
 *  The result of calling this function with <tt>iter == NULL</tt> is undefined.
 * @return This function returns true if @a iter refers to the end of its list instead of a valid list node.
 *  It returns false if the iterator refers to a valid list node.
 */
static inline bool drl_is_eol(const struct drl_iterator *const iter)
{
    return !iter->node;
}

/**
 * Add payload as a new node at a specific location of a list.
 *
 * The runtime complexity of this function is O(1).
 *
 * @param iter Points to an iterator.
 *  The new node is inserted between the iterator's node and its predecessor.
 *  If @a iter refers to the end of the list (i.e., when drl_is_eol(iter) returns @c true), the new node is inserted as the last node in the list.
 *  This also works for inserting the first node into an empty list.
 *  The result of calling this function with <tt>iter == NULL</tt> is undefined.
 * @param payload Points to the payload to add to the list.
 *  @a payload may be @c NULL, in which case a node with a @c NULL payload is prepended to the list.
 *  The same @a payload may be added to a list several times.
 *  In that case however, one needs to be particularly careful about when to de-allocate the payload data.
 *  Otherwise, one list node may still point to payload that has already been de-allocated for another list node.
 * @return Returns true if the new list node was successfully inserted into the list before @a iter.
 *  Returns false, if the list was not modified because the meta-data for the new node could not be allocated.
 */
bool drl_insert_before ARGS((struct drl_iterator *const iter,
                             void *const payload));

/**
 * Delete a node from a list.
 *
 * This function deletes the node that the given iterator refers to.
 *
 * @warning After drl_del() returns, the iterator @a iter is in an inconsistent state.
 * To make the iterator consistent again, call drl_next() or drl_prev().
 * The result of passing an inconsistent iterator to any other RList function is undefined.
 * Note that when drl_del() is invoked from a drl_map_fn() call-back function, drl_next() is invoked implicitely after the call-back returns.
 * Thus the call-back does not need to invoke drl_next() or drl_prev() itself.
 *
 * The runtime complexity of this function is O(1).
 *
 * @param iter Points to an iterator object.
 *  The iterator may not be uninitialized and must have previously been returned by one of the RList functions.
 *  The result of calling this function with <tt>iter == NULL</tt> is undefined.
 * @return If the list is associated with an rl_free_fn() call-back function, that call-back in invoked for the payload of the node to delete and this function returns @c NULL.
 *  Note that if in this case the payload is present more than once in the list, it is indeed de-allocated but other list nodes with the same payload remain in the list and still point to the now de-allocated payload memory.
 *  Thus, inserting a payload more than once into a list should be avoided.
 *  If the list is not associated with an rl_free_fn() call-back function, the payload is not de-allocated and this function returns the payload of the deleted node.
 *  Note that the payload itself may be @c NULL.
 */
void *drl_del ARGS((struct drl_iterator *const iter));

/**
 * Delete all nodes from a list.
 *
 * This is a convenience function which calls drl_del() for all nodes of a list.
 *
 * The runtime complexity of this function is O(n).
 *
 * @param list Points to the list object from which to delete all nodes.
 *  The result of calling this function with <tt>list == NULL</tt> is undefined.
 */
void drl_del_all ARGS((struct drl_list *const list));

/**
 * Find a list node based on a match call-back function.
 *
 * This function iterates linearly over a list, starting at the given iterator position, and applies the given matcher call-back function to each list node.
 * It returns an iterator for the first matching node.
 *
 * The average runtime complexity of this function is O(n/2) + O(match_fn) when @a iter refers to the first list node.
 *
 * @param iter Points to an iterator object.
 *  The iterator indicates the list to search and the list node to begin the search at.
 *  The result of calling this function with <tt>iter == NULL</tt> or an uninitialized iterator object is undefined.
 * @param match_fn Points to a call-back function that identifies the list node to return.
 *  The result of calling this function with <tt>match_fn == NULL</tt> is undefined.
 * @param match_context A context object passed to @a match_fn on every invocation as its second parameter.
 *  The semantics of @a match_context depend solely on the implementation of the @a match_fn call-back.
 *  @a match_context may be @c NULL if @a match_fn supports that.
 * @return If @a match_fn matches the node referred to by @a iter or a subsequent node, this function updates @a iter so that it refers to the first such matching node and returns @a iter.
 *  If @a match_fn matches no node in @a list, this function updates @a iter so that it refers to the end of the list and returns @a iter.
 *  If @a iter refers to the end of the list, this function does not update @a iter and returns @a iter.
 *  This function never returns NULL.
 */
struct drl_iterator *drl_find ARGS((struct drl_iterator *const iter,
                                    rl_match_fn match_fn,
                                    void *const match_context));

/**
 * Find a list node with a certain payload.
 *
 * This is a convenience function implemented via drl_find().
 *
 * The average runtime complexity of this function is O(n/2) when @a iter refers to the first list node.
 *
 * @param iter Points to an iterator object.
 *  The iterator indicates the list to search and the list node to begin the search at.
 *  The result of calling this function with <tt>iter == NULL</tt> or an uninitialized iterator object is undefined.
 * @param payload points to the payload to find in @a list.
 *  If @a payload is @c NULL, the first list node with a @c NULL payload is searched for.
 * @return If @a payload matches the node referred to by @a iter or a subsequent node, this function updates @a iter so that it refers to the first such matching node and returns @a iter.
 *  If @a match_fn matches no node in @a list, this function updates @a iter so that it refers to the end of the list and returns @a iter.
 *  If @a iter refers to the end of the list, this function does not update @a iter and returns @a iter.
 *  This function never returns NULL.
 */
struct drl_iterator *drl_find_payload ARGS((struct drl_iterator *const iter,
                                            void *const payload));

/**
 * Applies a call-back function to some or all nodes of a list.
 *
 * This function linearly iterates over a list, starting at the given iterator.
 * For each node, it first checks if an optional rl_match_fn() call-back matches the node.
 * If the node is a match or no rl_match_fn() call-back is supplied, this function invokes the drl_map_fn() call-back for the node.
 * It continues until the end of the list is reached or until the drl_map_fn() call-back returns a non-zero integer.
 *
 * The average runtime complexity of this function is O(n/2) + O(match_fn) + O(map_fn) when @a iter refers to the first list node.
 *
 * @param iter Points to an iterator object.
 *  The iterator indicates the list to map @a map_fn onto and it specifies the list node to begin the map operation at.
 *  The result of calling this function with <tt>iter == NULL</tt> or an uninitialized iterator object is undefined.
 * @param map_fn points to the call-back function to invoke on every node matching @a match_fn.
 *  The result of calling this function with <tt>map_fn == NULL</tt> is undefined.
 * @param map_context A context object passed to @a map_fn on every invocation as its second parameter.
 *  The semantics of @a map_context depend solely on the implementation of the @a map_fn call-back.
 *  @a map_context may be @c NULL if @a map_fn supports that.
 * @param match_fn points to the call-back function that identifies the subset of nodes that @a map_fn is invoked for.
 *  If @a match_fn is @c NULL, @a map_fn is invoked for all list nodes.
 * @param match_context A context object passed to @a match_fn on every invocation as its second parameter.
 *  The semantics of @a match_context depend solely on the implementation of the @a match_fn call-back.
 *  @a match_context may be @c NULL if @a match_fn supports that.
 * @return This function returns the value returned by the last invocation of @a map_fn().
 *  If @a map_fn() is not invoked because @a iter did not point to any remaining list nodes or because @a match_fn() did not return true for any list node, this function returns 0.
 * @sa drl_map()
 */
int drl_map ARGS((struct drl_iterator *const iter,
                  drl_map_fn map_fn,
                  void *const map_context,
                  rl_match_fn match_fn,
                  void *const match_context));

C_DECLARATIONS_END

#endif /* RLIST_DOUBLE_H */
