/*
 * Copyright (c) 2011, Stefan Götz <stefan.goetz@web.de>
 * 
 * Permission to use, copy, modify, and/or distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
 * REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
 * INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
 * LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
 * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
 * PERFORMANCE OF THIS SOFTWARE.
 */

/**
 * @internal
 * @file
 * RList implementation of singly-linked lists.
 */

#include <stdlib.h>
#include <assert.h>

#include "rlist/common.h"
#include "rlist/single.h"

/**
 * @internal
 * Determines the state an iterator is in.
 *
 * @param iter Points to the iterator object to check.
 *  The result of calling this function with <tt>iter == NULL</tt> is undefined.
 * @return This function returns the state as an @c rl_iterator_state_ value.
 */
/* This check is fairly unintuitive.
 * This is a classic memory/computation trade-off: either store the state explicitely in an iterator or derive it from the node/prev/list state indirectly as done here.
 * We go for computation so as not to bloat iterator objects. */
static inline enum rl_iterator_state_ srl_get_state_(const struct srl_iterator *const iter)
{
    assert(iter);
    assert(iter->list);

    if (iter->node) {
        if (iter->node != iter->prev) {
            return RL_ISTATE_IN_LIST_;
        } else {
            return RL_ISTATE_POST_DELETE_;
        }
    } else {
        if (iter->prev || !iter->list->head) {
            return RL_ISTATE_END_OF_LIST_;
        } else {
            return RL_ISTATE_POST_DELETE_;
        }
    }
}

bool srl_check_list_(const struct srl_list *const list)
{
    assert(list);
    assert(list->node_count != 0 || !list->head);
    assert(list->node_count == 0 ||  list->head);
    assert(list->node_count != 1 ||  (list->head && !list->head->next));
    assert(list->node_count == 1 || !(list->head && !list->head->next));
    assert(list->node_count <= 1 ||  (list->head &&  list->head->next));
    assert(list->node_count  > 1 || !(list->head &&  list->head->next));

    return list != NULL;
}

bool srl_check_iter_in_list_(const struct srl_iterator *const iter)
{
    /* is the iterator a valid object? */
    assert(iter != NULL);
    /* is the iterator's list consistent? */
    srl_check_list_(iter->list);
    /* is the current node a valid object? */
    assert(iter->node != NULL);
    /* the current and the previous node must not be the same */
    assert(iter->node != iter->prev);

    /* if prev is NULL, then node must be the first node in the list */
    assert(iter->prev != NULL || iter->node == iter->list->head);
    /* if node is the first node in the list, then prev must be NULL */
    assert(iter->node != iter->list->head || iter->prev == NULL);

    /* if prev is not NULL, then node must not be the first node in the list */
    assert(iter->prev == NULL || iter->node != iter->list->head);
    /* if node is not the first node in the list, then prev must not be NULL */
    assert(iter->node == iter->list->head || iter->prev != NULL);

    /* if prev is not NULL, then prev->next must be equal to node */
    assert(iter->prev == NULL || iter->prev->next == iter->node);

    return iter != NULL;
}

bool srl_check_iter_post_delete_(const struct srl_iterator *const iter)
{
    /* is the iterator a valid object? */
    assert(iter != NULL);
    /* is the iterator's list consistent? */
    srl_check_list_(iter->list);
    /* is node == prev to indicate post-delete? */
    assert(iter->node == iter->prev);

    return iter != NULL;
}

bool srl_check_iter_end_of_list_(const struct srl_iterator *const iter)
{
    /* is the iterator a valid object? */
    assert(iter != NULL);
    /* is the iterator's list consistent? */
    srl_check_list_(iter->list);
    /* is the current node NULL to indicate EOL? */
    assert(iter->node == NULL);

    /* if prev is NULL, then the list must be empty */
    assert(iter->prev != NULL || iter->list->head == NULL);
    /* if the list is empty, then prev must be NULL */
    assert(iter->list->head != NULL || iter->prev == NULL);

    /* if prev is not NULL, then it must be the last node in the list */
    assert(iter->prev == NULL || iter->prev->next == NULL);

    return iter != NULL;
}

bool srl_check_iter_in_list_or_end_of_list_(const struct srl_iterator *const iter)
{
    assert((srl_get_state_(iter) == RL_ISTATE_IN_LIST_  && srl_check_iter_in_list_(iter)) ||
           (srl_get_state_(iter) == RL_ISTATE_END_OF_LIST_ && srl_check_iter_end_of_list_(iter)));

    return iter != NULL;
}

struct srl_iterator *srl_next(struct srl_iterator *const iter)
{
    const enum rl_iterator_state_ state = srl_get_state_(iter);

    switch (state) {
    case RL_ISTATE_IN_LIST_:
        assert(srl_check_iter_in_list_(iter));
        /* advance iterator normally */
        iter->prev = iter->node;
        iter->node = iter->prev->next;
        break;
    case RL_ISTATE_POST_DELETE_:
        assert(srl_check_iter_post_delete_(iter));
        /* this branch implies iter->node == iter->prev */
        if (iter->prev) {
            /* with a good prev pointer, take its next pointer */
            iter->node = iter->prev->next;
        } else {
            /* with node and prev both NULL, take the first list node */
            /* srl_get_state_() should make sure that if the list was empty, the
             * iterator state is reported as EOL instead of PD */
            assert(iter->list->head);
            /* prev shall remain NULL because node is the first list node and
             * thus has no predecessor */
            iter->node = iter->list->head;
        }
        break;
    case RL_ISTATE_END_OF_LIST_:
        break;
    }

    if (srl_is_eol(iter)) {
        assert(srl_check_iter_end_of_list_(iter));
    } else {
        assert(srl_check_iter_in_list_(iter));
    }

    return iter;
}

bool srl_insert_before(struct srl_iterator *const iter,
                       void *const payload)
{
    assert(srl_check_iter_in_list_or_end_of_list_(iter));

    struct srl_node_ *const new_node = malloc(sizeof(*new_node));
    if (new_node) {
        *new_node = (struct srl_node_) { payload, iter->node };
        /* pointers that need to be updated:
         * list head, next pointer of previous node, prev pointer of iterator */
        if (iter->list->head == iter->node) {
            iter->list->head = new_node;
        }
        if (iter->prev) {
            iter->prev->next = new_node;
        }
        iter->prev              = new_node;
        iter->list->node_count += 1;

        assert(srl_check_iter_in_list_or_end_of_list_(iter));

        return true;
    }

    return false;
}

void *srl_del(struct srl_iterator *const iter)
{
    void            *ret = NULL;
    struct srl_node_ *next;

    assert(srl_check_iter_in_list_(iter));

    /* free or return the payload data */
    if (iter->list->free_fn) {
        iter->list->free_fn(iter->node->payload);
    } else {
        ret = iter->node->payload;
    }

    /* there are three pointers in our datastructures that may refer to
     * iter->node: 1) the list head pointer, 2) the previous node's next
     * pointer, and 3) the iterator's node pointer.
     * All three need to be set to a new value because iter->node is going to be
     * de-allocated. */
    next = iter->node->next;

    /* 1 of 3 */
    if (iter->list->head == iter->node) {
        iter->list->head = next;
    }
    /* 2 of 3 */
    if (iter->prev) {
        iter->prev->next = next;
    }
    free(iter->node);
    /* 3 of 3 */
    /* Set the iterator state to post-delete and let srl_next() fix it up */
    iter->node = iter->prev;

    iter->list->node_count -= 1;

    if (iter->list->head) {
        assert(srl_check_iter_post_delete_(iter));
    } else {
        assert(srl_check_iter_end_of_list_(iter));
    }

    return ret;
}

void srl_del_all(struct srl_list *const list)
{
    struct srl_iterator i;

    assert(srl_check_list_(list));

    while (!srl_is_eol(srl_first(list, &i))) {
        srl_del(&i);
    }

    assert(srl_check_list_(list));
}

/**
 * @internal
 * A srl_map_fn() call-back function which copies the iterator of the first list node it encounters to @a context and terminates the srl_map() iteration.
 *
 * @param iter points to an iterator that refers to a list node.
 * @param context assumed to point to an srl_iterator object.
 *  After calling this function, the data in @a iter is copied to @a context.
 * @return Returns false.
 */
static int rl_map_return_first_match_(__attribute__ ((unused)) struct srl_iterator *const iter,
                                      __attribute__ ((unused)) void *const context)
{
    return 1;
}

struct srl_iterator *srl_find(struct srl_iterator *const iter,
                              rl_match_fn match_fn,
                              void *const match_context)
{
    assert(srl_check_iter_in_list_or_end_of_list_(iter));
    assert(match_fn);

    srl_map(iter, rl_map_return_first_match_, NULL, match_fn, match_context);

    assert(srl_check_iter_in_list_or_end_of_list_(iter));

    return iter;
}

/**
 * @internal
 * Match a node by its payload.
 *
 * @param payload points to the payload of the node to check.
 * @param context points to the payload to be found in a node.
 * @return true when @a payload and @a context are equal.
 */
static bool rl_match_payload_(void *const payload,
                              void *const context)
{
    return payload == context;
}

struct srl_iterator *srl_find_payload(struct srl_iterator *const iter,
                                      void *const payload)
{
    return srl_find(iter, rl_match_payload_, payload);
}

int srl_map(struct srl_iterator *const iter,
            srl_map_fn map_fn,
            void *const map_context,
            rl_match_fn match_fn,
            void *const match_context)
{
    int ret = 0;

    assert(srl_check_iter_in_list_or_end_of_list_(iter));
    assert(map_fn);

    for (; !srl_is_eol(iter); srl_next(iter)) {
        if (!match_fn || match_fn(srl_payload(iter), match_context)) {
            ret = map_fn(iter, map_context);
            if (ret != 0) {
                break;
            }
        }
    }

    assert(srl_check_iter_in_list_or_end_of_list_(iter));

    return ret;
}
